; 32 bit Galois LFSR PRNG
; Copyright 2015 Johnathan Roatch
;
; Copying and distribution of this file, with or without
; modification, are permitted in any medium without royalty provided
; the copyright notice and this notice are preserved in all source
; code copies.  This file is offered as-is, without any warranty.
;
;
; Most old PRNG are some form of LFSR includeing CRCs.
; https://en.wikipedia.org/wiki/Linear_feedback_shift_register
;
; A fine LFSR to use is Greg Cook's CRC-16
; http://www.6502.org/source/integers/crc-more.html
;
; but I wanted a fast 32-bit LFSR so...
; http://forums.nesdev.com/viewtopic.php?f=2&t=9598
; http://forums.nesdev.com/viewtopic.php?f=10&t=11241
;
; The goal here is to select a large maximum-length polynomial
; that has as few taps as possible. I also choose a Galois
; LFSRs instead of Fibonacci LFSRs so that I could use a
; polynomial that will only xor 1 byte per bit.
;
; for a 32 bit state, there are three such polynomials.
; x^32 + x^30 + x^26 + x^25 + x^0
; x^32 + x^30 + x^28 + x^27 + x^26 + x^25 + x^0
; x^32 + x^31 + x^30 + x^29 + x^27 + x^25 + x^0
;
; The first one has the fewest taps. This is important for the
; optimized byte routine as it shifts and XORs the input byte
; for each tap.
;
; Finaly I flipped the bits from being right shifted to left
; shifted, So that the resulting low byte will automaticaly
; be in the accumulator on exit. Just like in:
; http://wiki.nesdev.com/w/index.php/Random_number_generator
;

.include "nes.inc"

.importzp rng_seed

.export rng_stepBit, rng_stepByte
.export rng_init

seed = rng_seed

.segment "CODE"

;;
; Initialize PRNG and Gather entropy from uninitialized OAM
;
; http://forums.nesdev.com/viewtopic.php?p=110200#p110200
.scope rng
init:
  txa
  pha
    ldx #$ff
    stx seed+0
    stx seed+1
    stx seed+2
    stx seed+3
    inx   ;,; ldx #$00
    stx OAM_ADDR
    @__loop:
      jsr rng_stepByte
      ;,; lda seed+0
      eor OAM_DATA
      ; also mix in zeropage in case OAM can't be read back.
      eor 0,x
      sta seed+0
      inx
    bne @__loop
    ; Reduce the entropy by 1 bit to avoid the all zero case.
    lda seed+3
    ora #%10000000
    sta seed+3
  pla
  tax
;,; jmp step_bit

;;
; Step the left shifted 32-bit Galois LFSR by 1 bit.
;
; Out A: Equals seed+0, bit 0 equals carry flag
; Out Carry: Output bit
step_bit:
  lda seed+0
  asl
  rol seed+1
  rol seed+2
  rol seed+3
  bcc @__skip
    eor #$c5
  @__skip:
  sta seed+0
rts

;;
; rng_stepByte is equivalent to calling rng_stepBit 8 times.
;
; Out A: Output byte.
; Out Y: Output High byte. (if subroutine is called twice)
step_byte:
  lda seed+3        ; Dealing with the low byte.
  lsr               ; start with x^30 term.
  lsr
  lsr
  lsr
  eor seed+3        ; x^26 term
  lsr
  eor seed+3        ; x^25 term
  lsr
  eor seed+0        ; XOR with original low byte.
  sta seed+0
  lda seed+3        ; Again for the high byte.
  asl               ; Start with x^25 term
  eor seed+3        ; x^26 term
  asl
  asl
  asl
  asl
  eor seed+3        ; x^30 term
  asl
  asl
  eor seed+3        ; x^32 term (or original high byte)
  ldy seed+2        ; finaly, rotate the bytes.
  sty seed+3
  ldy seed+1
  sty seed+2
  ldy seed+0
  sty seed+1
  sta seed+0
rts
.endscope

rng_init = rng::init
rng_stepByte = rng::step_byte
rng_stepBit = rng::step_bit
